"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_never_1 = __importDefault(require("assert-never"));
const chalk_1 = __importDefault(require("chalk"));
const http_1 = __importDefault(require("http"));
const buffer_1 = require("./buffer");
const matcher_1 = require("./matcher");
const persistence_1 = require("./persistence");
const sender_1 = require("./sender");
/**
 * A server that proxies or replays requests depending on the mode.
 */
class RecordReplayServer {
    constructor(options) {
        this.currentTapeRecords = [];
        this.replayedTapes = new Set();
        this.currentTapeRecords = [];
        this.mode = options.initialMode;
        this.proxiedHost = options.host;
        this.timeout = options.timeout || 5000;
        this.loggingEnabled = options.enableLogging || false;
        this.persistence = new persistence_1.Persistence(options.tapeDir);
        this.loadTape(DEFAULT_TAPE);
        this.server = http_1.default.createServer((req, res) => __awaiter(this, void 0, void 0, function* () {
            if (!req.url) {
                if (this.loggingEnabled) {
                    console.error(chalk_1.default.red("Received a request without URL."));
                }
                return;
            }
            if (!req.method) {
                if (this.loggingEnabled) {
                    console.error(chalk_1.default.red("Received a request without HTTP method."));
                }
                return;
            }
            try {
                const requestBody = yield receiveRequestBody(req);
                const requestPath = extractPath(req.url);
                if (requestPath === "/__proxay" ||
                    requestPath.startsWith("/__proxay/")) {
                    this.handleProxayApi(req.method, requestPath, requestBody, res);
                    return;
                }
                let record;
                switch (this.mode) {
                    case "replay":
                        record = matcher_1.findNextRecordToReplay(matcher_1.findRecordMatches(this.currentTapeRecords, req.method, requestPath, req.headers, requestBody), this.replayedTapes);
                        if (record) {
                            this.replayedTapes.add(record);
                            if (this.loggingEnabled) {
                                console.log(`Replayed: ${req.method} ${requestPath}`);
                            }
                        }
                        else {
                            if (this.loggingEnabled) {
                                console.warn(chalk_1.default.yellow(`Unexpected request ${req.method} ${requestPath} has no matching record in tapes.`));
                            }
                        }
                        break;
                    case "record":
                        record = yield this.proxy(req.method, requestPath, req.headers, requestBody);
                        this.addRecordToTape(record);
                        if (this.loggingEnabled) {
                            console.log(`Recorded: ${req.method} ${requestPath}`);
                        }
                        break;
                    case "mimic":
                        record = matcher_1.findNextRecordToReplay(matcher_1.findRecordMatches(this.currentTapeRecords, req.method, requestPath, req.headers, requestBody), this.replayedTapes);
                        if (record) {
                            this.replayedTapes.add(record);
                            if (this.loggingEnabled) {
                                console.log(`Replayed: ${req.method} ${requestPath}`);
                            }
                        }
                        else {
                            record = yield this.proxy(req.method, requestPath, req.headers, requestBody);
                            this.addRecordToTape(record);
                            if (this.loggingEnabled) {
                                console.log(`Recorded: ${req.method} ${requestPath}`);
                            }
                        }
                        break;
                    case "passthrough":
                        record = yield this.proxy(req.method, requestPath, req.headers, requestBody);
                        if (this.loggingEnabled) {
                            console.log(`Proxied: ${req.method} ${requestPath}`);
                        }
                        break;
                    default:
                        throw assert_never_1.default(this.mode);
                }
                if (record) {
                    this.sendResponse(record, res);
                }
                else {
                    res.statusCode = 500;
                    res.end();
                }
            }
            catch (e) {
                if (this.loggingEnabled) {
                    console.error(chalk_1.default.red("Unexpected error:"), e);
                }
                res.statusCode = 500;
                res.end();
            }
        }));
    }
    /**
     * Starts the server.
     */
    start(port) {
        return __awaiter(this, void 0, void 0, function* () {
            yield new Promise(resolve => this.server.listen(port, resolve));
        });
    }
    /**
     * Stops the server.
     */
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            yield new Promise(resolve => this.server.close(resolve));
        });
    }
    /**
     * Proxies a specific request and returns the resulting record.
     */
    proxy(requestMethod, requestPath, requestHeaders, requestBody) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.proxiedHost) {
                throw new Error("Missing proxied host");
            }
            try {
                return yield sender_1.send(this.proxiedHost, requestMethod, requestPath, requestHeaders, requestBody, this.timeout);
            }
            catch (e) {
                if (e.code) {
                    if (this.loggingEnabled) {
                        console.error(chalk_1.default.red(`Could not proxy request ${requestMethod} ${requestPath} (${e.code})`));
                    }
                }
                else {
                    if (this.loggingEnabled) {
                        console.error(chalk_1.default.red(`Could not proxy request ${requestMethod} ${requestPath}`, e));
                    }
                }
                throw e;
            }
        });
    }
    /**
     * Handles requests that are intended for Proxay itself.
     */
    handleProxayApi(requestMethod, requestPath, requestBody, res) {
        // Sending a request to /__proxay will return a 200 (so tests can identify whether
        // their backend is Proxay or not).
        if (requestMethod.toLowerCase() === "get" && requestPath === "/__proxay") {
            res.end("Proxay!");
        }
        // Sending a request to /__proxay/tape will pick a specific tape and/or a new mode.
        if (requestMethod.toLowerCase() === "post" &&
            requestPath === "/__proxay/tape") {
            const json = requestBody.toString("utf8");
            let tape;
            let mode;
            try {
                const body = JSON.parse(json);
                tape = body.tape || this.currentTape;
                mode = body.mode || this.mode;
            }
            catch (_a) {
                tape = null;
                mode = this.mode;
            }
            if (mode !== this.mode) {
                this.mode = mode;
            }
            if (tape) {
                if (!this.persistence.isTapeNameValid(tape)) {
                    const errorMessage = `Invalid tape name: ${tape}`;
                    if (this.loggingEnabled) {
                        console.error(chalk_1.default.red(errorMessage));
                    }
                    res.statusCode = 403;
                    res.end(errorMessage);
                    return;
                }
                if (this.loadTape(tape)) {
                    res.end(`Updated tape: ${tape}`);
                }
                else {
                    res.statusCode = 404;
                    res.end(`Missing tape: ${tape}`);
                }
            }
            else {
                this.unloadTape();
                res.end(`Unloaded tape`);
            }
        }
    }
    /**
     * Loads a specific tape into memory (erasing it in record mode).
     *
     * @returns Whether the tape was found or not (always true in record/mimic mode).
     */
    loadTape(tapeName) {
        this.currentTape = tapeName;
        if (this.loggingEnabled) {
            console.log(chalk_1.default.blueBright(`Loaded tape: ${tapeName}`));
        }
        switch (this.mode) {
            case "record":
                this.currentTapeRecords = [];
                this.persistence.saveTapeToDisk(this.currentTape, []);
                return true;
            case "replay":
                try {
                    this.currentTapeRecords = this.persistence.loadTapeFromDisk(this.currentTape);
                    return true;
                }
                catch (e) {
                    if (this.loggingEnabled) {
                        console.warn(chalk_1.default.yellow(e.message));
                    }
                    return false;
                }
            case "mimic":
                try {
                    this.currentTapeRecords = this.persistence.loadTapeFromDisk(this.currentTape);
                }
                catch (e) {
                    this.currentTapeRecords = [];
                    this.persistence.saveTapeToDisk(this.currentTape, []);
                }
                return true;
            case "passthrough":
                // Do nothing.
                return true;
            default:
                throw assert_never_1.default(this.mode);
        }
    }
    /**
     * Unloads the current tape, falling back to the default.
     */
    unloadTape() {
        this.loadTape(DEFAULT_TAPE);
    }
    /**
     * Adds a new record to the current tape and saves to disk.
     */
    addRecordToTape(record) {
        this.currentTapeRecords.push(record);
        this.persistence.saveTapeToDisk(this.currentTape, this.currentTapeRecords);
    }
    /**
     * Sends a particular response to the client.
     */
    sendResponse(record, res) {
        res.statusCode = record.response.status.code;
        Object.keys(record.response.headers).forEach(headerName => {
            const headerValue = record.response.headers[headerName];
            if (headerValue) {
                res.setHeader(headerName, headerValue);
            }
        });
        res.end(record.response.body);
    }
}
exports.RecordReplayServer = RecordReplayServer;
function receiveRequestBody(req) {
    const requestChunks = [];
    req.on("data", chunk => {
        requestChunks.push(buffer_1.ensureBuffer(chunk));
    });
    return new Promise(resolve => {
        req.on("end", () => resolve(Buffer.concat(requestChunks)));
    });
}
function extractPath(url) {
    const schemePosition = url.indexOf("://");
    let path;
    if (schemePosition !== -1) {
        const pathPosition = url.indexOf("/", schemePosition + 3);
        path = url.substr(pathPosition);
    }
    else {
        path = url;
    }
    return path;
}
const DEFAULT_TAPE = "default";
